export CMeanGaussian

"""
    CMeanGaussian{AbstractVar}(mapping, σ2, xlength)

Conditional Gaussian that maps an input z to a mean μx. The variance σ2 is
shared for all datapoints.  The mapping must output dimensions appropriate for
the chosen variance type:

# Arguments
- `mapping`: maps condition z to μ=mapping(z) (e.g. a Flux Chain)
- `σ2`: shared variance for all datapoints
- `xlength`: length of mean/variance vectors. Only needed for `ScalarVar`
- `AbstractVar`: one of the variance types: DiagVar, ScalarVar

# Example
```julia-repl
julia> p = CMeanGaussian{DiagVar}(Dense(2,3),ones(Float32,3))
CMeanGaussian{DiagVar}(mapping=Dense(2, 3), σ2=3-element Array{Float32,1}

julia> mean_var(p,ones(2))
(Float32[1.8698889, -0.24418116, 0.76614076], Float32[1.0, 1.0, 1.0])

julia> rand(p, ones(2))
3-element Array{Float32,2}:
 0.1829532154926673
 0.1235498922955946
 0.0767166501426535
```
"""
struct CMeanGaussian{V<:AbstractVar,M,S<:AbstractVector} <: AbstractCGaussian
    mapping::M
    σ::S
    xlength::Int
    _nograd::Dict{Symbol,Bool}
end

CMeanGaussian{V}(m::M, σ::S, xlength::Int, d::Dict{Symbol,Bool}) where {V,M,S} =
    CMeanGaussian{V,M,S}(m,σ,xlength,d)

function CMeanGaussian{V}(m::M, σ::AbstractVector, xlength::Int) where {V,M}
    @assert _eltype(m) == eltype(σ)
    _nograd = Dict(:σ => σ isa NoGradArray)
    σ = _nograd[:σ] ? σ.data : σ
    CMeanGaussian{V}(m, σ, xlength, _nograd)
end

CMeanGaussian{ScalarVar}(m::M, σ::Number, xlength::Int) where M =
    CMeanGaussian{ScalarVar}(m, [σ], xlength)

CMeanGaussian{DiagVar}(m, σ) = CMeanGaussian{DiagVar}(m, σ, size(σ,1))

mean(p::CMeanGaussian, z::AbstractArray) = p.mapping(z)

function var(p::CMeanGaussian{DiagVar}, z::AbstractArray)
    T = eltype(p.σ)
    σ2 = p.σ .* p.σ .+ T(1e-8)
    σ2 * fill!(similar(σ2, 1, size(z,2)), 1)
end

function var(p::CMeanGaussian{ScalarVar}, z::AbstractArray)
    T = eltype(p.σ)
    σ2 = p.σ .* p.σ .+ T(1e-8)
    σ2 .* fill!(similar(p.σ, p.xlength, size(z,2)), 1)
end

cov(p::CMeanGaussian, z::AbstractArray) = Diagonal(var(p,z))
mean_var(p::CMeanGaussian, z::AbstractArray) = (mean(p, z), var(p, z))
length(p::CMeanGaussian) = p.xlength
_eltype(m) = eltype(first(Flux.params(m)))
eltype(p::CMeanGaussian) = _eltype(p.mapping)


# make sure that parameteric constructor is called...
function Flux.functor(p::CMeanGaussian{V,M,S}) where {V,M,S}
    fs = fieldnames(typeof(p))
    nt = (; (name=>getfield(p, name) for name in fs)...)
    nt, y -> CMeanGaussian{V}(y...)
end

function Flux.trainable(p::CMeanGaussian)
    ps = [getfield(p,k) for k in keys(p._nograd) if !p._nograd[k]]
    (p.mapping, ps...)
end

function Base.show(io::IO, p::CMeanGaussian{V}) where V
    e = repr(p.mapping)
    e = sizeof(e)>50 ? "($(e[1:47])...)" : e
    m = "CMeanGaussian{$V}(mapping=$e, σ2=$(summary(p.σ))"
    print(io, m)
end
